/*
 * This is the phytec peb loader kernel module.
 */
#define DEBUG

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/peb_interface.h>

#include <linux/i2c.h>
#include <linux/spi/spi.h>
#include <linux/irq.h>
#include <linux/gpio.h>

#define NUMSLOTS	2
extern struct peb_interfaces_data peb_interfaces; /* Exported by platformcode*/

enum peb_type {
	NO_PEB = 0,
	PEB_A001,
	PEB_A003,
	PEB_EEPROM,
};

enum peb_bus_type {
	NO_BUS = 0,
	SPI,
	I2C,
};

static struct {
	char			*param[16];
	unsigned int   		num_params;
	unsigned short  	i2c_addr;
	enum peb_bus_type	bus;
	int			type;
	struct spi_master 	*spi_master;
	struct spi_device	*spi_device;
	struct i2c_adapter 	*i2c_adapter;	
	struct i2c_client	*i2c_device;
	int (*add)(unsigned int slot_no);
} slot[NUMSLOTS];

module_param_array_named(slot1, slot[0].param, charp, &slot[0].num_params, S_IRUGO);
MODULE_PARM_DESC(slot1, "peb name, i2c address" 
	"Example: \"slot1=PEB-A-001 or slot1=PEB-A-003,0x40\"");
module_param_array_named(slot2, slot[1].param, charp, &slot[1].num_params, S_IRUGO);
MODULE_PARM_DESC(slot2, "peb name, i2c address"
	"Example: \"slot2=PEB-A-001 or slot2=PEB-A-003,0x40\"");

static struct spi_board_info peb_spi_board_info;
static struct i2c_board_info peb_i2c_devices_board;

/* PEB-A-001 */
#include <linux/can/platform/mcp251x.h>
#if !defined(CONFIG_CAN_MCP251X) && !defined(CONFIG_CAN_MCP251X_MODULE)
	#error Need MCP251X CAN Support in kernel
#endif

static struct mcp251x_platform_data mcp251x_info[NUMSLOTS] = {
	{
		.oscillator_frequency = 24000000,
		.power_enable = NULL,
		.transceiver_enable = NULL,
		.board_specific_setup = NULL,
	},{
		.oscillator_frequency = 24000000,
		.power_enable = NULL,
		.transceiver_enable = NULL,
		.board_specific_setup = NULL,
	}
};

int add_mcp251x(unsigned int slot_no)
{
	struct peb_slot_data peb_slot = peb_interfaces.peb_slots[slot_no];
	struct spi_master *master = slot[slot_no].spi_master;
	struct spi_device *spi_device = slot[slot_no].spi_device;

	pr_debug("%s on slot %d\n",__func__,slot_no + 1);

	strcpy(peb_spi_board_info.modalias, "mcp251x");
	peb_spi_board_info.max_speed_hz = 2*1000*1000;
	peb_spi_board_info.irq = peb_slot.irq; 
	peb_spi_board_info.chip_select = peb_slot.spi_cs;
	peb_spi_board_info.bus_num = peb_slot.spi_busnum;
	peb_spi_board_info.platform_data = &mcp251x_info[slot_no];

	spi_device = spi_new_device(master, &peb_spi_board_info);

	if(!spi_device){
		pr_err("%s Adding mcp251x failed\n",__func__);
		return -EBUSY;
	}
	pr_debug("%s spi_device %p registered\n",__func__,spi_device);
	return 0;
}

/* PEB-A-003 */
#include <linux/spi/max7301.h>
#if !defined(CONFIG_GPIO_MAX7300) && !defined(CONFIG_GPIO_MAX7300_MODULE)
	#error Need MAX7300 Support in kernel
#endif

#define MAXINPUTGPIOS	8
#define MAXOUTPUTGPIOS	8
#define MAXGPIOS	(MAXINPUTGPIOS + MAXOUTPUTGPIOS)

struct max7301_platform_data peb_max7300[NUMSLOTS] = {
	{
		.base = -1, 
		.input_pullup_active = 0,
	},{
		.base = -1, 
		.input_pullup_active = 0,
	}
};

struct max7301 *max7301_drvdata[NUMSLOTS];
static const char *names[MAXGPIOS] = {
	"DIG_INP1", "DIG_INP2", "DIG_INP3", "DIG_INP4",	
	"DIG_INP5", "DIG_INP6", "DIG_INP7", "DIG_INP8",	
	"DIG_OUT1", "DIG_OUT2", "DIG_OUT3", "DIG_OUT4",	
	"DIG_OUT5", "DIG_OUT6", "DIG_OUT7", "DIG_OUT8",	
};

int add_max7301(unsigned int slot_no)
{
	struct i2c_adapter *adapter= slot[slot_no].i2c_adapter;
	struct i2c_client *i2c_device = slot[slot_no].i2c_device;

	struct max7301 *ts = max7301_drvdata[slot_no];
	int i;

	pr_debug("%s on slot %d\n",__func__,slot_no + 1);

	strcpy(peb_i2c_devices_board.type,"max7300");
	peb_i2c_devices_board.platform_data = &peb_max7300[slot_no];

	peb_i2c_devices_board.addr = 
		(slot[slot_no].i2c_addr)?slot[slot_no].i2c_addr:0x40;

	i2c_device = i2c_new_device(adapter, &peb_i2c_devices_board);
	if (!i2c_device)
		return -ENODEV;

	pr_debug("%s i2c_device %p registered\n",__func__,i2c_device);

	ts = dev_get_drvdata(&i2c_device->dev);
	if (!ts) {
		pr_err("%s module max7300 loaded? (modprobe max7300)\n",__func__);
		goto error;
	}

	ts->chip.names = names;
	//FIXME 2 PEBs

	for (i = 0; i < MAXINPUTGPIOS ; i++) {
		gpio_request(ts->chip.base + i, names[i]);
		gpio_direction_input(ts->chip.base + i);
		gpio_export(ts->chip.base + i, 0);
	}
	for (i = 0; i <  MAXOUTPUTGPIOS ; i++) {
		gpio_request(ts->chip.base + MAXINPUTGPIOS + i, names[i + MAXINPUTGPIOS]);
		gpio_direction_output(ts->chip.base + MAXINPUTGPIOS + i, 0);
		gpio_export(ts->chip.base + MAXINPUTGPIOS + i, 0);
	}

	return 0;
error:
	i2c_unregister_device(i2c_device);
	pr_debug("%s i2c_device %p unregistered\n",__func__,i2c_device);
	return -ENODEV;

}
/* PEB_EEPROM */
#include <linux/spi/eeprom.h>

static struct spi_eeprom peb_at25[NUMSLOTS] = {
	{
		.name		= "at25320an",
		.byte_len	= 4096,
		.page_size	= 32,
		.flags		= EE_ADDR2,
	},{
		.name		= "at25320an",
		.byte_len	= 4096,
		.page_size	= 32,
		.flags		= EE_ADDR2,
	}
};

int add_eeprom(unsigned int slot_no)
{
	struct peb_slot_data peb_slot = peb_interfaces.peb_slots[slot_no];
	struct spi_master *master = slot[slot_no].spi_master;
	struct spi_device *spi_device = slot[slot_no].spi_device;
	pr_debug("%s on slot %d\n",__func__,slot_no + 1);

	strcpy(peb_spi_board_info.modalias, "at25");
	peb_spi_board_info.max_speed_hz = 30000;
	peb_spi_board_info.irq = peb_slot.irq; 
	peb_spi_board_info.chip_select = peb_slot.spi_cs;
	peb_spi_board_info.bus_num = peb_slot.spi_busnum;
	peb_spi_board_info.platform_data = &peb_at25[slot_no];

	spi_device = spi_new_device(master, &peb_spi_board_info);

	if(!spi_device){
		pr_err("%s Adding spi eeprom mcp251x failed\n",__func__);
		return -EBUSY;
	}
	pr_debug("%s spi_device %p registered\n",__func__,spi_device);
	return 0;
}

/* Common */
static struct {
	char			*name;
	enum peb_type		type;
	enum peb_bus_type	bus;
	int (*add)(unsigned int slot_no);
} peb_names[] = {
	{"PEB-A-001", PEB_A001, SPI, &add_mcp251x},
	{"PEB-A-003", PEB_A003, I2C, &add_max7301},
	{"PEB-EEPROM", PEB_EEPROM, SPI, &add_eeprom},
};

/*
 * peb_init - function to insert the platformcode for peb cards
 *
 * Returns 0 if successfull, otherwise -ENODEV
 */

static int __init peb_init(void)
{
	int i, j;
	int i2c_busnum;
	u16 spi_busnum;

	if ((slot[0].num_params == 0) && (slot[1].num_params == 0)) {
		pr_err("%s slots not defined!\n", __func__);
		pr_err("%s Example: \"slot1=PEB-A-001 slot2=PEB-A-003,0x40\"\n",
					 __func__);
		return -EINVAL;
	}

	for (i=0; i < NUMSLOTS; i++) {
		if (!slot[i].num_params || !slot[i].param[0]) 
			continue;

		for (j = 0; j < ARRAY_SIZE(peb_names); j++)
			if (!strcmp(peb_names[j].name, slot[i].param[0])) {
				slot[i].type = peb_names[j].type;
				slot[i].bus = peb_names[j].bus;
				slot[i].add = peb_names[j].add;
				break;
			}

		if (j == ARRAY_SIZE(peb_names)) {
			pr_err("%s Unkown PEB type on slot%d\n",__func__, i + 1);
			return -ENODEV;
		}

		pr_debug("%s %s (%d) on slot%d uses %s%s\n", __func__, slot[i].param[0], slot[i].type,
				i + 1, (slot[i].bus == I2C)?"i2c":"", (slot[i].bus == SPI)?"spi":"");

		switch (slot[i].bus)  {
			case (I2C):
				if(slot[i].num_params > 1)
					slot[i].i2c_addr = simple_strtoul(slot[i].param[1], NULL, 16);
				else
					slot[i].i2c_addr = 0;
				i2c_busnum = peb_interfaces.peb_slots[i].i2c_busnum;
				slot[i].i2c_adapter = i2c_get_adapter(i2c_busnum);
				if (!slot[i].i2c_adapter) {
					pr_err("%s I2C BUS %d not found\n", __func__, i2c_busnum);
					return -ENODEV;
				}
				if (slot[i].add(i))
					return -ENODEV;
				break;
			case (SPI):
				spi_busnum = peb_interfaces.peb_slots[i].spi_busnum;
				slot[i].spi_master = spi_busnum_to_master(spi_busnum);
				if(!slot[i].spi_master) {
					pr_err("%s SPI BUS %d not found\n",__func__, spi_busnum);
					return -ENODEV;
				}
				if (slot[i].add(i))
					return -ENODEV;
				break;
			case (NO_BUS):
			default:
				pr_err("%s Unkown Bus Type\n",__func__);
				return -ENODEV;
		}
	}
	return 0;
}

/*
 * peb_cleanup - function to cleanup the 
 */

static void __exit peb_cleanup(void)
{
	int i ,j;
	for (i=0; i < NUMSLOTS; i++) {
		if (slot[i].spi_device) {
			spi_unregister_device(slot[i].spi_device);
			pr_debug("%s slot %d: spi_device %p unregistered\n", 
				__func__,i + 1, slot[i].spi_device);
		}
		if (slot[i].i2c_device) {
			i2c_unregister_device(slot[i].i2c_device);
			pr_debug("%s slot %d: i2c_device %p unregistered\n",
				__func__,i + 1, slot[i].i2c_device);
		}
		if (max7301_drvdata[i]) 
			for (j = 0; j <  MAXGPIOS ; j++)
				gpio_free(max7301_drvdata[i]->chip.base + j);
	}
}

module_init(peb_init);
module_exit(peb_cleanup);

MODULE_DESCRIPTION("Register PEB PhytecExpansionBoards");
MODULE_LICENSE("GPL");

